/*
 * File: rho.c
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 * MACOS port by Aria Stewart <aredridel@nbtsc.org>
 */

/***

The original MicroEMACS code exploits string arrays and #defines to
create primitive dictionary structures.  The problem is that you can
only declare the array once.  So all references to that array end up
in one file.  This file uses keytab[] (rho_key.h) and names[]
(rho_name.h) in this way.

***/


#include	<ctype.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#if MACOS
#include	<time.h>
#endif                          /* MACOS */
#include	"rho.h"
#include	"bind.h"
#include	"buffer.h"
#include	"cmode.h"
#include	"cursor.h"
#include	"display.h"
#include	"envars.h"
#include	"exec.h"
#include	"file.h"
#include	"fileio.h"
#include	"isearch.h"
#include	"killring.h"
#include	"line.h"
#include	"macros.h"
#include	"messages.h"
#include	"process.h"
#include	"region.h"
#include	"replace.h"
#include	"search.h"
#include	"tags.h"
#include	"ufuncs.h"
#include	"terminal.h"
#include	"utils.h"
#include	"window.h"
#include	"word.h"
#include	"rho_init.h"    /*  var initialization and externs         */
#include	"rho_name.h"    /*  name_string-to-function binding table  */
#include	"rho_key.h"     /*  key_code-to-function binding table     */

extern time_t time(time_t *);

/*++

Execute the startup file.  Can be macros, C error-processing file
or the default RHO_RC

++*/
static int startup(sfname)
char *sfname;                   /* name of startup file (null if default) */

{
    char *fname;                /* resulting file name to execute */
    char name[NSTRING];         /* name with extention */

    /* init TERM */
    callout(SHELLINIT);

    /* look up the startup file */
    if (*sfname != 0) {
        strcpy(name, sfname);
        fname = search_paths(name, TRUE);
    } else
        fname = search_paths((char *) RHO_RC, TRUE);

    /* if it isn't around, don't sweat it */
    if (fname == NULL)
        return (TRUE);

    /* otherwise, execute it */
    return (do_file(fname));
}



/*++

Process a command line.   May be called any time.

++*/
void docmdline(argc, argv)
int argc;
char *argv[];

{

    BUFFER *bp;                 /* temp buffer pointer */
    BUFFER *firstbp = NULL;     /* ptr to first buffer in cmd line */
    char bname[LENPATH];        /* buffer name of file to read */
    int carg;                   /* current arg to scan */
    int firstfile;              /* first file flag */
    int gchar;                  /* and what character? */
    int gline;                  /* if so, what line? */
    int gotoflag;               /* do we need to goto a line at start? */
    int searchflag;             /* Do we need to search at start? */

    exec_error = FALSE;         /* no macro error pending */
    firstfile = TRUE;           /* no file to edit yet */
    gotoflag = FALSE;           /* set to off to begin with */
    searchflag = FALSE;         /* set to off to begin with */

    gline = 1;
    gchar = 1;                  /* line and character to go to */
    lterm[0] = 0;               /* standard line terminators */

    startup("");                /* source ~/.rhorc */

    for (carg = 1; carg < argc; ++carg) {       /* Parse command line */

        /* switches */
        if (argv[carg][0] == '-') {
            switch (argv[carg][1]) {
            case 'g':          /* -g for initial goto line */
            case 'G':
                gotoflag = TRUE;
                gline = str2int(&argv[carg][2]);
                break;
            case 's':          /* -s for initial regex search string */
            case 'S':
                searchflag = TRUE;
                strncpy((char *) pat, &argv[carg][2], NPAT);
                setjtable();
                break;
            default:
                break;          /* bad switch */
            }

        } else if (argv[carg][0] == '+') {      /* +<line num> for cscope */

            gotoflag = TRUE;
            gline = str2int(&argv[carg][1]);

        } else {                /* Process input files */
//            if (fexist(argv[carg])) {
            strcpy(bname, argv[carg]);
            bp = exist_buffer(bname, TRUE, 0);
            bp->b_active = FALSE;
            if (firstfile) {
                firstbp = bp;
                firstfile = FALSE;
            }
//            }
        }

    }                           /* end for loop */

    /* if files to read, read the firstfile */
    bp = exist_buffer(mainbuf, FALSE, 0);
    if (firstfile == FALSE && (gflags & GFREAD)) {
        make_bp_top(firstbp);
        curbp->b_mode |= gmode;
        update(TRUE);
        mlwrite(lastmesg);
    } else
        bp->b_mode |= gmode;

    /* goto or search on firstfile? */
    if (gotoflag) {
        if ((goto_line(TRUE, gline) == FALSE) ||
            (next_character(TRUE, gchar - 1) == FALSE)) {
            update(FALSE);
            mlwrite(TEXT102);   /*"[Bogus goto argument]" */
        }
    } else if (searchflag) {
        if (search_again(FALSE, 0) == FALSE)
            update(FALSE);
    }

}



/*++

This is the general command execution routine. It handles the fake binding
of all the keys to "self-insert". It also clears out the "thisflag" word,
and arranges to move it to the "lastflag", so that the next command can
look at it. Return the status of command.

++*/
int execute(c, f, n)
int c;                          /* key to execute */
int f;                          /* prefix argument flag */
int n;                          /* prefix value */

{
    int status;
    KEYTAB *key;                /* key entry to execute */

    /* if the keystroke is a bound function...do it */
    key = exist_binding(c);
    if (key != NULL) {

        /* Don't reset the function type flags on a prefix */
        if ((key->k_type == BINDFNC) &&
            ((key->k_fp == meta_prefix) || (key->k_fp == ctlx)))
            status = execute_key(key, f, n);
        else {
            thisflag = 0;
            status = execute_key(key, f, n);
            lastflag = thisflag;
        }

        return (status);
    }

    /*
     * If a space was typed, fill column is defined, the argument is non-
     * negative, wrap mode is enabled, and we are now past fill column,
     * and we are not read-only, perform word wrap.
     */
    if (c == ' ' && (curwp->w_bufp->b_mode & MDWRAP) && fillcol > 0 &&
        n >= 0 && getccol(FALSE) > fillcol &&
        (curwp->w_bufp->b_mode & MDVIEW) == FALSE)
        execute_key(&wraphook, FALSE, 1);

    if ((c >= 0x20 && c <= 0xFF)) {     /* Self inserting.      */
        if (n <= 0) {           /* Fenceposts.          */
            lastflag = 0;
            return (n < 0 ? FALSE : TRUE);
        }
        thisflag = 0;           /* For the future.      */

        /* replace or overwrite mode, not at the end of a string */
        if (curwp->w_bufp->b_mode & (MDREPL | MDOVER) &&
            curwp->w_doto < lused(curwp->w_dotp)) {

            /* if we are in replace mode, or
               (next char is not a tab or we are at a tab stop) */
            if (curwp->w_bufp->b_mode & MDREPL ||
                (lgetc(curwp->w_dotp, curwp->w_doto) != '\t' ||
                 getccol(FALSE) % hardtab == (hardtab - 1)))
                deletenchars(1L, FALSE);
        }

        /* do the appropriate insertion */
        if (c == '}' && (curbp->b_mode & MDCMOD) != 0)
            status = insbrace(n, c);
        else if (c == '#' && (curbp->b_mode & MDCMOD) != 0)
            status = inspound();
        else
            status = insertnchars(n, c);

        /* check auto-save mode */
        if (curbp->b_mode & MDASAVE)
            if (--asavecount == 0) {
                /* and save the file if needed */
                update_screen(FALSE, 0);
                save_file(FALSE, 0);
                asavecount = asave;
            }

        lastflag = thisflag;
        return (status);
    }
    TTbeep();
    mlwrite(TEXT19);            /*"[Key not bound]" */
    lastflag = 0;               /* Fake last flags. */
    return (FALSE);
}




/*++

This is called to let the user edit something.	Note that if you
arrange to be able to call this from a macro, you will have
invented the "recursive-edit" function.

++*/
int editloop()
{
    int c;                      /* command character */
    int f;                      /* default flag */
    int n;                      /* numeric repeat count */
    int mflag;                  /* negative flag on repeat */
    int basec;                  /* c stripped of meta character */
    int oldflag;                /* old last flag value */
    char tmp[LENPATH];          /* for lines 310ff */

    /* load buffers waiting for files (mult. files on cmd line) */
    if ((curbp->b_name[0] == '?') && (curbp->b_name[1] == ' ')
        && (curbp->b_name[strlen(curbp->b_name) - 1] == '/')) {
        for (c = 2; c <= (int) strlen(curbp->b_name); ++c)
            tmp[c - 2] = curbp->b_name[c];
        strcpy(curbp->b_name, tmp);
        discmd = TRUE;
        find_file(0, 0);
        if ((curbp->b_name[0] == ':') && (curbp->b_name[1] == ' ')
            && (curbp->b_name[strlen(curbp->b_name) - 1] == '/')) {
            strcpy(curbp->b_name, mainbuf);
            update_modeline();
        }
    }

    /* setup to process commands */
    lastflag = 0;               /* Fake last flags.     */

  loop:
    /* if a macro error is pending, wait for a character */
    if (exec_error) {
        mlforce(TEXT227);       /*"\n--- Press any key to Continue ---" */
        tgetc();
        sgarbf = TRUE;
        update(FALSE);
        mlferase();
        exec_error = FALSE;
    }

    /* if we were called as a subroutine and want to leave, do so */
    if (eexitflag)
        return (eexitval);

    /* execute the "command" macro...normally null */
    oldflag = lastflag;         /* preserve lastflag through this */
    execute_key(&cmdhook, FALSE, 1);
    lastflag = oldflag;

    /* update position on current modeline? */
    if (posflag)
        if (!type_ahead())
            update_modeline();

    /* Fix up the screen    */
    update(FALSE);

    /* get the next command from the keyboard */
    discmd = TRUE;
    disinp = TRUE;
    c = mlgetkey();

    /* if there is something on the command line, clear it */
    if (mpresf != FALSE) {
        mlerase();
        update(FALSE);
    }

    /* override the arguments if prefixed */
    if (prefix) {
        if (islower(c & 255))
            c = (c & ~255) | toupper(c & 255);
        c |= prefix;
        f = predef;
        n = prenum;
        prefix = 0;
    } else {
        f = FALSE;
        n = 1;
    }

    /* do META-# processing if needed */

    basec = c & ~META;          /* strip meta char off if there */
    if ((c & META) && ((basec >= '0' && basec <= '9') || basec == '-') &&
        (exist_binding(c) == NULL)) {
        f = TRUE;               /* there is a # arg */
        n = 0;                  /* start with a zero default */
        mflag = 1;              /* current minus flag */
        c = basec;              /* strip the META */
        while ((c >= '0' && c <= '9') || (c == '-')) {
            if (c == '-') {
                /* already hit a minus or digit? */
                if ((mflag == -1) || (n != 0))
                    break;
                mflag = -1;
            } else {
                n = n * 10 + (c - '0');
            }
            if ((n == 0) && (mflag == -1))      /* lonely - */
                mlwrite("Arg:");
            else
                mlwrite("Arg: %d", n * mflag);

            c = mlgetkey();     /* get the next key */
        }
        n = n * mflag;          /* figure in the sign */
    }

    /* do ^U repeat argument processing */

    if (c == reptc) {           /* ^U, start argument   */
        f = TRUE;
        n = 4;                  /* with argument of 4 */
        mflag = 0;              /* that can be discarded. */
        mlwrite("Arg: 4");
        while ((((c = mlgetkey()) >= '0') && (c <= '9')) || c == reptc
               || c == '-') {
            if (c == reptc)
                if ((n > 0) == ((n * 4) > 0))
                    n = n * 4;
                else
                    n = 1;
            /*
             * If dash, and start of argument string, set arg.
             * to -1.  Otherwise, insert it.
             */
            else if (c == '-') {
                if (mflag)
                    break;
                n = 0;
                mflag = -1;
            }
            /*
             * If first digit entered, replace previous argument
             * with digit and set sign.  Otherwise, append to arg.
             */
            else {
                if (!mflag) {
                    n = 0;
                    mflag = 1;
                }
                n = 10 * n + c - '0';
            }
            mlwrite("Arg: %d", (mflag >= 0) ? n : (n ? -n : -1));
        }
        /*
         * Make arguments preceded by a minus sign negative and change
         * the special argument "^U -" to an effective "^U -1".
         */
        if (mflag == -1) {
            if (n == 0)
                n++;
            n = -n;
        }
    }

    /* and execute the command */
    execute(c, f, n);
    goto loop;
}



/*++

Initialize all of the buffers, windows and screens. The buffer name is
passed down as an argument, because the main routine may have been told
to read in a file by default, and we want the buffer name to be right.

++*/
void edinit(bname)
char bname[];                   /* name of buffer to initialize */

{
    BUFFER *bp;
    int index;

    /* init the kill ring */
    for (index = 0; index < NRING; index++) {
        kbufp[index] = (KILL *) NULL;
        kbufh[index] = (KILL *) NULL;
        kskip[index] = 0;
        kused[index] = KBLOCK;
    }
    kill_index = 0;

    /* initialize some important globals */

    readhook.k_fp = nop;        /* set internal hooks to OFF */
    readhook.k_type = BINDFNC;
    wraphook.k_fp = wrap_word;
    wraphook.k_type = BINDFNC;
    cmdhook.k_fp = nop;
    cmdhook.k_type = BINDFNC;
    writehook.k_fp = nop;
    writehook.k_type = BINDFNC;
    bufhook.k_fp = nop;
    bufhook.k_type = BINDFNC;
    exbhook.k_fp = nop;
    exbhook.k_type = BINDFNC;

    /* allocate the first buffer */
    bp = exist_buffer(bname, TRUE, 0);  /* First buffer */
    blistp = exist_buffer("[Buffers]", TRUE, BFINVS);
    if (bp == NULL || blistp == NULL)
        exit_status(1);

    /* and allocate the default screen */
    first_screen = (SCRN *) NULL;
    (void) init_screen("MAIN", bp);
    if (first_screen == (SCRN *) NULL)
        exit_status(1);

    /* set the current default screen/buffer/window */
    curbp = bp;
    curwp = wheadp = first_screen->s_cur_window =
        first_screen->s_first_window;
}



/*++

Look through the list of buffers. Return TRUE if there are any changed
buffers. Buffers that hold internal stuff are not considered; Return
FALSE if no buffers have been changed.  Helper for main.exit_rho

++*/
int pendingsaves()
{
    BUFFER *bp;

    bp = bheadp;
    while (bp != NULL) {
        if (strcmp(bp->b_name, mainbuf) == 0) {
            bp = bp->b_bufp;
            continue;
        }
        if ((bp->b_flag & BFINVS) == 0 && (bp->b_flag & BFCHG) != 0)
            return (TRUE);
        bp = bp->b_bufp;
    }
    return (FALSE);
}



/*++

Quit command. If an argument, always exit_rho. Otherwise confirm if a buffer
has been changed and not written out. 

++*/
int exit_rho(f, n)
int f, n;                       /* prefix flag and argument */

{
    int status;                 /* return status */

    if (f != FALSE              /* Argument forces it.  */
        || pendingsaves() == FALSE      /* All buffers clean or */
        || (status = mlyesno(TEXT104)) == TRUE) {       /*"Modified buffers exist. Leave anyway" */
        if (f)
            status = exit_status(n);
        else
            status = exit_status(0);
    }
    if (system(NULL) != 0)
        system("rm _rho_* 2>/dev/null");        /* remove any dangling tmpfiles */
    mlerase();
    return (status);
}


/*++

Internal exit with status.

++*/
int exit_status(status)
int status;                     /* return status of rho */
{
    eexitflag = TRUE;
    eexitval = status;

    return (TRUE);
}




/*++

Change the editor mode status.

++*/
int adjustmode(kind, global)
int kind;                       /* true = set,          false = delete */
int global;                     /* true = global flag,  false = current buffer flag */
{
    char *scan;                 /* scanning pointer to convert prompt */
    int i;                      /* loop index */
    int status;                 /* error return on input */
    char prompt[50];            /* string to prompt user with */
    char mdbuf[NPAT];           /* buffer to recieve mode name into */

    /* build the proper prompt string */
    if (global)
        strcpy(prompt, TEXT62); /*"Global mode to " */
    else
        strcpy(prompt, TEXT63); /*"Mode to " */

    if (kind == TRUE)
        strcat(prompt, TEXT64); /*"add: " */
    else
        strcat(prompt, TEXT65); /*"delete: " */

    /* prompt the user and get an answer */

    status = mlreply(prompt, mdbuf, NPAT - 1);
    if (status != TRUE)
        return (status);

    /* make it uppercase */

    scan = mdbuf;
    while (*scan)
        toupperp((unsigned char *) scan++);

    /* test it against the modes we know */

    for (i = 0; i < NUMMODES; i++) {
        if (strcmp(mdbuf, modename[i]) == 0) {
            /* finding a match, we process it */
            if (kind == TRUE)
                if (global) {
                    gmode |= (1 << i);
                    if ((1 << i) == MDOVER)
                        gmode &= ~MDREPL;
                    else if ((1 << i) == MDREPL)
                        gmode &= ~MDOVER;
                } else {
                    curbp->b_mode |= (1 << i);
                    if ((1 << i) == MDOVER)
                        curbp->b_mode &= ~MDREPL;
                    else if ((1 << i) == MDREPL)
                        curbp->b_mode &= ~MDOVER;
            } else if (global)
                gmode &= ~(1 << i);
            else
                curbp->b_mode &= ~(1 << i);
            /* display new mode line */
            if (global == 0)
                update_modeline();
            mlerase();
            return (TRUE);
        }
    }

    mlwrite(TEXT66);            /*"No such mode!" */
    return (FALSE);
}




/*++

Prompt and set an editor mode.

++*/
int setmod(f, n)
int f, n;
{

    f = n = 0;                  /* splint */

    return (adjustmode(TRUE, FALSE));
}



/*++

Prompt and delete an editor mode.

++*/
int delmode(f, n)
int f, n;
{

    f = n = 0;                  /* splint */

    return (adjustmode(FALSE, FALSE));
}


/*++

Prompt and set a global editor mode

++*/
int setgmode(f, n)
int f, n;
{

    f = n = 0;                  /* splint */

    return (adjustmode(TRUE, TRUE));
}


/*++

Prompt and delete a global editor mode.

++*/
int delgmode(f, n)
int f, n;
{

    f = n = 0;                  /* splint */

    return (adjustmode(FALSE, TRUE));
}


/*++

This is the primary entry point that is used by command line invocation.

++*/
int main(argc, argv)
int argc;                       /* # of arguments */
char *argv[];                   /* argument strings */

{
    int status = 0;

    eexitflag = FALSE;
    vtinit();                   /* Terminal */
    if (eexitflag) {
        vttidy();
        exit(status);
    } else {
        edinit(mainbuf);        /* Buffers, windows, screens */
        varinit();              /* User variables */
        docmdline(argc, argv);  /* Process the command line */
        status = editloop();    /* Main loop   */
        vttidy();               /* Cleanup terminal   */
        exit(status);
    }

}
